import bs
import bsUtils
import random
import bsVector

_maps = {}

def preloadPreviewMedia():
    bs.getModel('levelSelectButtonOpaque')
    bs.getModel('levelSelectButtonTransparent')
    for m in _maps.values(): m.getPreviewTexture()
    
def registerMap(m):
    """ Register a level class with the game. """
    if _maps.has_key(m.name):
        raise Exception("map \"" + m.name + "\" already registered")
    _maps[m.name] = m

def getFilteredMapName(name):
    """ filters a map name to account for name changes, etc so old configs still work """
    # some legacy name fallbacks...
    if name == 'AlwaysLand' or name == 'Happy Land': name = 'Happy Thoughts'
    if name == 'Hockey Arena': name = 'Hockey Stadium'
    return name

def getTranslatedMapName(name):
    try: translated = bs.getResource('translations.mapsNames')[name]
    except Exception: translated = None
    if translated is None: translated = name
    return translated
    
def getMapList(gameType):
    """ return a list of maps by game type """
    maps = [m[0] for m in _maps.items() if gameType in m[1].gameTypes]
    maps.sort()
    return maps

def getMapClass(name):
    """ return a map type given a name """
    name = getFilteredMapName(name)
    try: return _maps[name]
    except Exception: raise Exception("Map not found: '"+name+"'")
    
class Map(bs.Actor):
    """ base class for maps """
    defs = None
    name = "Map"
    gameTypes = []
    preloaded = False
    @classmethod
    def preload(self):
        """ Preload map media.
        This runs the class's doPreload function if need be to get it ready to run.
        Preloading can be fired for a soon-to-be-needed map to speed its creation.
        This is a classmethod since it is not run on map instances but rather on
        the class as a whole before instances are made"""
        if not self.preloaded:
            self.doPreload()
            self.preloaded = True

    @classmethod
    def getPreviewTextureName(self):
        return None

    @classmethod
    def getPreviewTexture(self):
        """ returns the preview texture for the map, or None if it has none """
        mapTex = self.getPreviewTextureName()
        if mapTex is not None: mapTex = bs.getTexture(mapTex)
        return mapTex

    @classmethod
    def doPreload(self):
        pass

    @classmethod
    def getMusicType(self):
        """ if this map has particular music associated with it, return it here. """
        return None

    def __init__(self,activity,retain):
        import _bs
        bs.Actor.__init__(self,activity,retain)
        self.preload()

        # set some defaults
        bs.globals.happyThoughts = 0
        bs.globals.floorReflection = False
        bs.globals.debrisFriction = 1.0
        bs.globals.debrisKillHeight = -50.0
        bs.globals.areaOfInterestBounds = self.getDefBoundBox("areaOfInterestBounds") or (-500,-500,-500,500,500,500)
        bs.globals.shadowOffset = (0,0,0)
        bs.globals.shadowScale = (1.0,1.0)
        bs.globals.shadowOrtho = False
        bs.globals.tint = (1.1,1.0,0.9)
        bs.globals.ambientColor = (1,1,1)
        bs.globals.vignetteOuter = (0.6,0.6,0.6)
        bs.globals.vignetteInner = (0.95,0.95,0.95)

        _bs._setMapBounds(self.getDefBoundBox("levelBounds") or (-30,-10,-30,30,100,30))

        try: bs.globals.shadowRange = [self.defs.points[v][1] for v in ['shadowLowerBottom','shadowLowerTop','shadowUpperBottom','shadowUpperTop']]
        except Exception: bs.globals.shadowRange = [-4,0,10,15]

        self.spawnPoints = self.getDefPoints("spawn") or [(0,0,0,0,0,0)]
        self.ffaSpawnPoints = self.getDefPoints("ffaSpawn") or [(0,0,0,0,0,0)]
        self.spawnByFlagPoints = self.getDefPoints("spawnByFlag") or [(0,0,0,0,0,0)]
        self.flagPoints = self.getDefPoints("flag") or [(0,0,0)]
        self.flagPointDefault = self.getDefPoint("flagDefault") or (0,1,0)
        self.powerupSpawnPoints = self.getDefPoints("powerupSpawn") or [(0,0,0)]
        self.tntPoints = self.getDefPoints("tnt") or []
        self.isHockey = False
        self.isFlying = False
        self._nextFFAStartIndex = 0

    def isPointNearEdge(self,p,running=False):
        " for bot purposes "
        return False

    def getDefBoundBox(self,name):
        """Returns a 6 member bounds tuple or None if it is not defined."""
        try:
            b = self.defs.boxes[name]
            return (b[0]-b[6]/2.0,b[1]-b[7]/2.0,b[2]-b[8]/2.0,
                    b[0]+b[6]/2.0,b[1]+b[7]/2.0,b[2]+b[8]/2.0);
        except Exception:
            return None
        
    def getDefPoint(self,name):
        """Return a single defined point or a default value in its absence."""
        try:
            return self.defs.points[name]
        except Exception:
            return None

    def getDefPoints(self,name):
        """Return a list of points - as many sequential ones are defined (flag1, flag2, flag3, etc)."""
        if self.defs and self.defs.points.has_key(name+"1"):
            pointList = []
            i = 1
            while self.defs.points.has_key(name+str(i)):
                p = self.defs.points[name+str(i)]
                if len(p) == 6:
                    pointList.append(p)
                else:
                    if len(p) != 3: raise Exception("invalid point")
                    pointList.append(p+(0,0,0))
                i += 1
            return pointList
        else:
            return None
        
    def getStartPosition(self,team):
        """Return a starting position in the map for the given team."""
        pt = self.spawnPoints[team%len(self.spawnPoints)]
        xRange = (-0.5,0.5) if pt[3] == 0 else (-pt[3],pt[3])
        zRange = (-0.5,0.5) if pt[5] == 0 else (-pt[5],pt[5])
        pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
        return pt

    def getFFAStartPosition(self,players):
        """Return a starting position in one of the FFA spawn areas."""

        # get positions for existing players
        playerPts = []
        for player in players:
            try:
                pt = bsVector.Vector(*player.actor.node.translate)
                # fresh players will have translate 0,0,0; need to fix this..
                if pt.gameData == [0.0,0.0,0.0]:
                    pt = bsVector.Vector(*player.actor._lastStandPos)
                playerPts.append(pt)
            except Exception: pass

        def _getPt():
            pt = self.ffaSpawnPoints[self._nextFFAStartIndex]
            self._nextFFAStartIndex = (self._nextFFAStartIndex+1)%len(self.ffaSpawnPoints)
            xRange = (-0.5,0.5) if pt[3] == 0 else (-pt[3],pt[3])
            zRange = (-0.5,0.5) if pt[5] == 0 else (-pt[5],pt[5])
            pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
            return pt

        if len(playerPts) == 0:
            return _getPt()
        else:
            # lets calc several start points and then pick whichever is farthest from all existing players
            farthestPtDist = -1.0
            farthestPt = None
            for i in range(10):
                testPt = bsVector.Vector(*_getPt())
                closestPlayerDist = 9999.0
                closestPlayerPt = None
                for pp in playerPts:
                    dist = (pp-testPt).length()
                    if dist < closestPlayerDist:
                        closestPlayerDist = dist
                        closestPlayerPt = pp
                if closestPlayerDist > farthestPtDist:
                    farthestPtDist = closestPlayerDist
                    farthestPt = testPt
            return tuple(farthestPt.data)

    def getFlagPosition(self,team):
        """Return a starting position in the map for the given flag."""
        if team == None:
            return self.flagPointDefault[:3]
        else:
            return self.flagPoints[team%len(self.flagPoints)][:3]

    def handleMessage(self,m):
        if isinstance(m,bs.HitMessage):
            if (m.srcNode != 0):
                pass
                #print 'FIXME do map kickback'
                # bs.nodeMessage(m.srcNode,"impulse",m.srcBody,m.pos[0],m.pos[1],m.pos[2],
                #                -0.65*m.force[0]*m.kickBack,-0.75*m.force[1]*m.kickBack,-0.65*m.force[2]*m.kickBack)

        #elif isinstance(m,bs.DieMessage): bs.deleteNode(self.node)
        elif isinstance(m,bs.DieMessage): self.node.delete()
        else: bs.Actor.handleMessage(self,m)

######### now go ahead and register some maps #########


class HockeyStadium(Map):
    import hockeyStadiumDefs as defs
    name = "Hockey Stadium"

    gameTypes = ['melee','hockey','teamFlag','keepAway']

    @classmethod
    def getPreviewTextureName(self):
        return 'hockeyStadiumPreview'

    @classmethod
    def doPreload(self):
        self.models = (bs.getModel("hockeyStadiumOuter"),
                      bs.getModel("hockeyStadiumInner"),
                      bs.getModel("hockeyStadiumStands"))
        self.collideModel = bs.getCollideModel("hockeyStadiumCollide")
        self.tex = bs.getTexture("hockeyStadium")
        self.standsTex = bs.getTexture("footballStadium")
        self.iceMaterial = bs._newMaterial()
        bs._materialAddAction(self.iceMaterial,actions=("partMod","friction",0.01))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={'model':self.models[0],
                                      'collideModel':self.collideModel,
                                      'colorTexture':self.tex,
                                      'materials':[bs.footingMaterial,self.iceMaterial]})
        self.floor = bs.newNode("terrain",
                                owner=self.node,
                                attrs={"model":self.models[1],
                                       "colorTexture":self.tex,
                                       "opacity":0.92,
                                       "opacityInLowOrMediumQuality":1.0,
                                       "materials":[bs.footingMaterial,self.iceMaterial]})
        self.stands = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.models[2],
                                        "visibleInReflections":False,
                                        "colorTexture":self.standsTex})

        bs.globals.floorReflection = True
        bs.globals.debrisFriction = 0.3
        bs.globals.debrisKillHeight = -0.3

        bs.globals.tint = (1.2,1.3,1.33)
        bs.globals.ambientColor = (1.15,1.25,1.6)
        bs.globals.vignetteOuter = (0.66,0.67,0.73)
        bs.globals.vignetteInner = (0.93,0.93,0.95)

        self.isHockey = True

registerMap(HockeyStadium)


class FootballStadium(Map):
    import footballStadiumDefs as defs
    name = "Football Stadium"

    gameTypes = ['melee','football','teamFlag','keepAway']

    @classmethod
    def getPreviewTextureName(self):
        return 'footballStadiumPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("footballStadium")
        self.collideModel = bs.getCollideModel("footballStadiumCollide")
        self.tex = bs.getTexture("footballStadium")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"model":self.model,
                                      "collideModel":self.collideModel,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        bs.globals.tint = (1.3,1.2,1.0)
        bs.globals.ambientColor = (1.3,1.2,1.0)
        bs.globals.vignetteOuter = (0.57,0.57,0.57)
        bs.globals.vignetteInner = (0.9,0.9,0.9)

    def isPointNearEdge(self,p,running=False):
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(FootballStadium)


class BridgitMap(Map):
    import bridgitLevelDefs as defs
    name = "Bridgit"
    gameTypes = ["melee","teamFlag",'keepAway']

    @classmethod
    def getPreviewTextureName(self):
        return 'bridgitPreview'

    @classmethod
    def doPreload(self):
        self.modelTop = bs.getModel("bridgitLevelTop")
        self.modelBottom = bs.getModel("bridgitLevelBottom")
        self.modelBG = bs.getModel("natureBackground")
        self.collideModel = bs.getCollideModel("bridgitLevelCollide")
        self.tex = bs.getTexture("bridgitLevelColor")
        self.modelBGTex = bs.getTexture("natureBackgroundColor")
        self.collideBG = bs.getCollideModel("natureBackgroundCollide")

        self.railingCollideModel = bs.getCollideModel("bridgitLevelRailingCollide")
    
        self.bgMaterial = bs._newMaterial()
        bs._materialAddAction(self.bgMaterial,actions=('partMod','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.modelTop,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        bs.globals.tint = (1.1,1.2,1.3)
        bs.globals.ambientColor = (1.1,1.2,1.3)
        bs.globals.vignetteOuter = (0.65,0.6,0.55)
        bs.globals.vignetteInner = (0.9,0.9,0.93)

registerMap(BridgitMap)


class BigGMap(Map):
    import bigGDefs as defs
    name = "Big G"
    gameTypes = ["race",'melee','keepAway','teamFlag','kingOfTheHill','conquest']

    @classmethod
    def getPreviewTextureName(self):
        return 'bigGPreview'

    @classmethod
    def doPreload(self):
        self.modelTop = bs.getModel("bigG")
        self.modelBottom = bs.getModel("bigGBottom")
        self.modelBG = bs.getModel("natureBackground")
        self.collideModel = bs.getCollideModel("bigGCollide")
        self.tex = bs.getTexture("bigG")
        self.modelBGTex = bs.getTexture("natureBackgroundColor")
        self.collideBG = bs.getCollideModel("natureBackgroundCollide")
        self.bumperCollideModel = bs.getCollideModel("bigGBumper")

    
        self.bgMaterial = bs._newMaterial()
        bs._materialAddAction(self.bgMaterial,actions=('partMod','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "color":(0.7,0.7,0.7),
                                      "model":self.modelTop,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "color":(0.7,0.7,0.7),
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.bumperCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        bs.globals.tint = (1.1,1.2,1.3)
        bs.globals.ambientColor = (1.1,1.2,1.3)
        bs.globals.vignetteOuter = (0.65,0.6,0.55)
        bs.globals.vignetteInner = (0.9,0.9,0.93)

registerMap(BigGMap)


class RoundaboutMap(Map):
    import roundaboutLevelDefs as defs
    name = "Roundabout"
    gameTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(self):
        return 'roundaboutPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("roundaboutLevel")
        self.modelBottom = bs.getModel("roundaboutLevelBottom")
        self.modelBG = bs.getModel("natureBackground")
        self.collideModel = bs.getCollideModel("roundaboutLevelCollide")
        self.tex = bs.getTexture("roundaboutLevelColor")
        self.modelBGTex = bs.getTexture("natureBackgroundColor")
        self.collideBG = bs.getCollideModel("natureBackgroundCollide")
    
        self.railingCollideModel = bs.getCollideModel("roundaboutLevelBumper")

        self.bgMaterial = bs._newMaterial()
        bs._materialAddAction(self.bgMaterial,actions=('partMod','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.modelBG,
                                    "lighting":False,
                                    "colorTexture":self.modelBGTex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (1.0,1.05,1.1)
        bs.globals.ambientColor = (1.0,1.05,1.1)
        bs.globals.shadowOrtho = True
        bs.globals.vignetteOuter = (0.63,0.65,0.7)
        bs.globals.vignetteInner = (0.97,0.95,0.93)

registerMap(RoundaboutMap)


class MonkeyFaceMap(Map):
    import monkeyFaceLevelDefs as defs
    name = "Monkey Face"
    gameTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(self):
        return 'monkeyFacePreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("monkeyFaceLevel")
        self.bottomModel = bs.getModel("monkeyFaceLevelBottom")
        self.modelBG = bs.getModel("natureBackground")
        self.collideModel = bs.getCollideModel("monkeyFaceLevelCollide")
        self.tex = bs.getTexture("monkeyFaceLevelColor")
        self.modelBGTex = bs.getTexture("natureBackgroundColor")
        self.collideBG = bs.getCollideModel("natureBackgroundCollide")
    
        self.railingCollideModel = bs.getCollideModel("monkeyFaceLevelBumper")

        self.bgMaterial = bs._newMaterial()
        bs._materialAddAction(self.bgMaterial,actions=('partMod','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (1.1,1.2,1.2)
        bs.globals.ambientColor = (1.2,1.3,1.3)
        bs.globals.vignetteOuter = (0.60,0.62,0.66)
        bs.globals.vignetteInner = (0.97,0.95,0.93)

registerMap(MonkeyFaceMap)


class ZigZagMap(Map):
    import zigZagLevelDefs as defs
    name = "Zigzag"
    gameTypes = ["melee","keepAway","teamFlag","conquest",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(self):
        return 'zigzagPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("zigZagLevel")
        self.modelBottom = bs.getModel("zigZagLevelBottom")
        self.modelBG = bs.getModel("natureBackground")
        self.collideModel = bs.getCollideModel("zigZagLevelCollide")
        self.tex = bs.getTexture("zigZagLevelColor")
        self.modelBGTex = bs.getTexture("natureBackgroundColor")
        self.collideBG = bs.getCollideModel("natureBackgroundCollide")

        self.railingCollideModel = bs.getCollideModel("zigZagLevelBumper")

        self.bgMaterial = bs._newMaterial()
        bs._materialAddAction(self.bgMaterial,actions=('partMod','friction',10.0))
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (1.0,1.15,1.15)
        bs.globals.ambientColor = (1.0,1.15,1.15)
        bs.globals.vignetteOuter = (0.57,0.59,0.63)
        bs.globals.vignetteInner = (0.97,0.95,0.93)

registerMap(ZigZagMap)


class ThePadMap(Map):
    import thePadLevelDefs as defs
    name = "The Pad"
    gameTypes = ["melee","keepAway","teamFlag",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(self):
        return 'thePadPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("thePadLevel")
        self.bottomModel = bs.getModel("thePadLevelBottom")
        self.collideModel = bs.getCollideModel("thePadLevelCollide")
        self.tex = bs.getTexture("thePadLevelColor")
        self.bgTex = bs.getTexture("menuBG")
        self.bgModel = bs.getModel("thePadBGSmall")
    
        self.railingCollideModel = bs.getCollideModel("thePadLevelBumper")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (1.1,1.1,1.0)
        bs.globals.ambientColor = (1.1,1.1,1.0)
        bs.globals.vignetteOuter = (0.7,0.65,0.75)
        bs.globals.vignetteInner = (0.95,0.95,0.93)

registerMap(ThePadMap)


class DoomShroomMap(Map):
    import doomShroomLevelDefs as defs
    name = "Doom Shroom"
    gameTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(self):
        return 'doomShroomPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("doomShroomLevel")
        self.collideModel = bs.getCollideModel("doomShroomLevelCollide")
        self.tex = bs.getTexture("doomShroomLevelColor")
        self.bgTex = bs.getTexture("doomShroomBGColor")
        self.bgModel = bs.getModel("doomShroomBG")
        self.stemModel = bs.getModel("doomShroomStem")
        self.collideBG = bs.getCollideModel("doomShroomStemCollide")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               name='Terrain Root',
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex},
                              name='Terrain BG')
        self.stem = bs.newNode("terrain",
                               owner=self.node,
                               attrs={"model":self.stemModel,
                                      "lighting":False,
                                      "colorTexture":self.tex},
                               name='Terrain Unlit')
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,bs.deathMaterial]},
                                    name='Terrain BG Collide')

        bs.globals.tint = (0.82,1.10,1.15)
        bs.globals.ambientColor = (0.9,1.3,1.1)
        bs.globals.shadowOrtho = False
        bs.globals.vignetteOuter = (0.76,0.76,0.76)
        bs.globals.vignetteInner = (0.95,0.95,0.99)

    def isPointNearEdge(self,p,running=False):
        x = p.x()
        z = p.z()
        xAdj = x*0.125
        zAdj = (z+3.7)*0.2
        if running:
            xAdj *= 1.4
            zAdj *= 1.4
        return (xAdj*xAdj+zAdj*zAdj > 1.0)

registerMap(DoomShroomMap)


class TipTopMap(Map):
    import tipTopLevelDefs as defs
    name = "Tip Top"
    gameTypes = ["melee","keepAway","teamFlag",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(self):
        return 'tipTopPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("tipTopLevel")
        self.bottomModel = bs.getModel("tipTopLevelBottom")
        self.collideModel = bs.getCollideModel("tipTopLevelCollide")
        self.tex = bs.getTexture("tipTopLevelColor")
        self.bgTex = bs.getTexture("tipTopBGColor")
        self.bgModel = bs.getModel("tipTopBG")

        self.railingCollideModel = bs.getCollideModel("tipTopLevelBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "color":(0.7,0.7,0.7),
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "color":(0.7,0.7,0.7),
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "color":(0.4,0.4,0.4),
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (0.8,0.9,1.3)
        bs.globals.ambientColor = (0.8,0.9,1.3)
        bs.globals.vignetteOuter = (0.79,0.79,0.69)
        bs.globals.vignetteInner = (0.97,0.97,0.99)

registerMap(TipTopMap)


class CragCastleMap(Map):
    import cragCastleDefs as defs
    name = "Crag Castle"
    gameTypes = ["melee","keepAway","teamFlag",'conquest']

    @classmethod
    def getPreviewTextureName(self):
        return 'cragCastlePreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("cragCastleLevel")
        self.bottomModel = bs.getModel("cragCastleLevelBottom")
        self.collideModel = bs.getCollideModel("cragCastleLevelCollide")
        self.tex = bs.getTexture("cragCastleLevelColor")
        self.bgTex = bs.getTexture("menuBG")
        self.bgModel = bs.getModel("thePadBG")

        self.railingCollideModel = bs.getCollideModel("cragCastleLevelBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bs.globals.shadowOrtho = True
        bs.globals.shadowOffset = (0,0,-5.0)
        bs.globals.tint = (1.15,1.05,0.75)
        bs.globals.ambientColor = (1.15,1.05,0.75)
        bs.globals.vignetteOuter = (0.6,0.65,0.6)
        bs.globals.vignetteInner = (0.95,0.95,0.95)

registerMap(CragCastleMap)



class TowerDMap(Map):
    import towerDLevelDefs as defs
    name = "Tower D"
    gameTypes = []

    @classmethod
    def getPreviewTextureName(self):
        return 'towerDPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("towerDLevel")
        self.collideModel = bs.getCollideModel("towerDLevelCollide")
        self.tex = bs.getTexture("towerDLevelColor")
        self.bgTex = bs.getTexture("menuBG")
        self.bgModel = bs.getModel("thePadBG")
        self.playerWallCollideModel = bs.getCollideModel("towerDPlayerWall")

        self.playerWallMaterial = bs._newMaterial()
        bs._materialAddAction(self.playerWallMaterial,actions=(("partMod","friction",0.0)))

        # anything that needs to hit the wall should apply this.
        self.collideWithWallMaterial = bs._newMaterial()
        bs._materialAddAction(self.playerWallMaterial,
                             conditions=("dstNotMaterial",self.collideWithWallMaterial),
                             actions=("partMod","collide",False))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.playerWall = bs.newNode("terrain",
                                     owner=self.node,
                                     attrs={"collideModel":self.playerWallCollideModel,
                                            "affectBGDynamics":False,
                                            "materials":[self.playerWallMaterial]})
        bs.globals.tint = (1.15,1.11,1.03)
        bs.globals.ambientColor = (1.2,1.1,1.0)
        bs.globals.vignetteOuter = (0.7,0.73,0.7)
        bs.globals.vignetteInner = (0.95,0.95,0.95)

    def isPointNearEdge(self,p,running=False):
        # see if we're within edgeBox
        boxes = self.defs.boxes
        boxTranslate = boxes['edgeBox'][0:3]
        boxScale = boxes['edgeBox'][6:9]
        boxTranslate2 = boxes['edgeBox2'][0:3]
        boxScale2 = boxes['edgeBox2'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        x2 = (p.x() - boxTranslate2[0])/boxScale2[0]
        z2 = (p.z() - boxTranslate2[2])/boxScale2[2]
        # if we're outside of *both* boxes we're near the edge
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5) and (x2 < -0.5 or x2 > 0.5 or z2 < -0.5 or z2 > 0.5)

registerMap(TowerDMap)


class AlwaysLandMap(Map):
    import alwaysLandLevelDefs as defs
    name = "Happy Thoughts"
    gameTypes = ["melee","keepAway","teamFlag","conquest",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(self):
        return 'alwaysLandPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("alwaysLandLevel")
        self.bottomModel = bs.getModel("alwaysLandLevelBottom")
        self.bgModel = bs.getModel("alwaysLandBG")
        self.collideModel = bs.getCollideModel("alwaysLandLevelCollide")
        self.tex = bs.getTexture("alwaysLandLevelColor")
        self.bgTex = bs.getTexture("alwaysLandBGColor")

    @classmethod
    def getMusicType(self):
        return 'Flying'
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex})

        bs.globals.happyThoughts = 1
        bs.globals.shadowOffset = (0.0,8.0,5.0)
        bs.globals.tint = (1.3,1.23,1.0)
        bs.globals.ambientColor = (1.3,1.23,1.0)
        bs.globals.vignetteOuter = (0.64,0.59,0.69)
        bs.globals.vignetteInner = (0.95,0.95,0.93)
        self.isFlying = True

        # throw out some tips on flying
        t = bs.newNode('text',
                       owner=self.getActivity(),
                       attrs={'text':bs.getResource('pressJumpToFlyText'),
                              'scale':[1.1],
                              'translate':(0,200),
                              'hAlign':'center',
                              'vAttach':'bottom'},name='Tip')
        c = bs.newNode("combine",owner=t,attrs={'input0':0.8,'input1':1.0,'input2':1.0})
        c.connectAttr('output',t,'color')
        bsUtils.animate(c,'input3',{0:0,1000:1,7000:1,8000:0})
        #bs.gameTimer(8000,bs.WeakCall(bs.deleteNode,t))
        bs.gameTimer(8000,t.delete)

        
registerMap(AlwaysLandMap)


class StepRightUpMap(Map):
    import stepRightUpLevelDefs as defs
    name = "Step Right Up"
    gameTypes = ["melee","keepAway","teamFlag","conquest"]


    @classmethod
    def getPreviewTextureName(self):
        return 'stepRightUpPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("stepRightUpLevel")
        self.collideModel = bs.getCollideModel("stepRightUpLevelCollide")
        self.tex = bs.getTexture("stepRightUpLevelColor")
        self.bgTex = bs.getTexture("menuBG")
        self.bgModel = bs.getModel("thePadBG")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})

        bs.globals.tint = (1.2,1.1,1.0)
        bs.globals.ambientColor = (1.2,1.1,1.0)
        bs.globals.vignetteOuter = (0.7,0.65,0.75)
        bs.globals.vignetteInner = (0.95,0.95,0.93)

registerMap(StepRightUpMap)


class CourtyardMap(Map):
    import courtyardLevelDefs as defs
    name = "Courtyard"
    gameTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(self):
        return 'courtyardPreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("courtyardLevel")
        self.collideModel = bs.getCollideModel("courtyardLevelCollide")
        self.tex = bs.getTexture("courtyardLevelColor")
        self.bgTex = bs.getTexture("menuBG")
        self.bgModel = bs.getModel("thePadBG")
        self.playerWallCollideModel = bs.getCollideModel("courtyardPlayerWall")
    
        self.playerWallMaterial = bs._newMaterial()
        bs._materialAddAction(self.playerWallMaterial,actions=(("partMod","friction",0.0)))

        # anything that needs to hit the wall should apply this.
        self.collideWithWallMaterial = bs._newMaterial()
        bs._materialAddAction(self.playerWallMaterial,
                             conditions=("dstNotMaterial",self.collideWithWallMaterial),
                             actions=("partMod","collide",False))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})

        # in challenge games, put up a wall to prevent players
        # from getting in the turrets (that would foil our brilliant AI)
        if 'CoopSession' in str(type(bs.getSession())):
            self.playerWall = bs.newNode("terrain",
                                         owner=self.node,
                                         attrs={"collideModel":self.playerWallCollideModel,
                                                "affectBGDynamics":False,
                                                "materials":[self.playerWallMaterial]})

        bs.globals.tint = (1.2,1.17,1.1)
        bs.globals.ambientColor = (1.2,1.17,1.1)
        bs.globals.vignetteOuter = (0.6,0.6,0.64)
        bs.globals.vignetteInner = (0.95,0.95,0.93)

    def isPointNearEdge(self,p,running=False):
        # count anything off our ground level as safe (for our platforms)
        #if p.y() > 3.1: return False
        # see if we're within edgeBox
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(CourtyardMap)


class RampageMap(Map):
    import rampageLevelDefs as defs
    name = "Rampage"
    gameTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(self):
        return 'rampagePreview'

    @classmethod
    def doPreload(self):
        self.model = bs.getModel("rampageLevel")
        self.bottomModel = bs.getModel("rampageLevelBottom")
        self.collideModel = bs.getCollideModel("rampageLevelCollide")
        self.tex = bs.getTexture("rampageLevelColor")
        self.bgTex = bs.getTexture("rampageBGColor")
        self.bgTex2 = bs.getTexture("rampageBGColor2")
        self.bgModel = bs.getModel("rampageBG")
        self.bgModel2 = bs.getModel("rampageBG2")

        self.railingCollideModel = bs.getCollideModel("rampageBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg2 = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel2,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex2})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})

        bs.globals.tint = (1.2,1.1,0.97)
        bs.globals.ambientColor = (1.3,1.2,1.03)
        bs.globals.vignetteOuter = (0.62,0.64,0.69)
        bs.globals.vignetteInner = (0.97,0.95,0.93)

    def isPointNearEdge(self,p,running=False):
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(RampageMap)
